#!/usr/bin/env python

from collections import defaultdict

from distributions import *
from schedcat.model.LLCLayout import *
from schedcat.model.LLC_params import *
import schedcat.model.ohc as ohc
import random
import time

from pulp import *
import gurobipy

import numpy as np

capped_wss = {'B' : [512, 512, 512, 512],
              'C' : 512}

def LPLLCways(ts_mcx, vartype, csch = bypass_scheme(), olap = True, debug = False, noIso = False):
#     print "vars:", len(ts_mcx), vartype, csch, olap, debug, noIso

    ts_mcx.setScheme(csch)
    ts_mcx.A = ts_mcx.level('A')
    ts_mcx.B = ts_mcx.level('B')
    ts_mcx.C = ts_mcx.level('C')
     
    #----------------------- Gather Data -----------------------------#
    UA_Adata = np.zeros((CORES, csch.mW+1))
    UA_Bdata = np.zeros((CORES, csch.mW+1))
    UA_Cdata = np.zeros((CORES, csch.mW+1))
    
    UB_Bdata = np.zeros((CORES, csch.mW+1))
    UB_Cdata = np.zeros((CORES, csch.mW+1))
    UC_Cdata = np.zeros(csch.mW+1)
    
    hdata = np.zeros(csch.mW+1)
    Hdata = np.zeros(csch.mW+1)
    
    minAperiods = []
    for core in range(0, 4):
        if len(ts_mcx.A.core(core)) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in ts_mcx.A.core(core)]))
    
    for w in range(0,csch.mW+1):
        wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
                      [area(w, csch.B[i].S) for i in range(0, CORES)],
                      area(w, csch.C.S), csch.mW, csch.mS)
                      
        if noIso:
            wsch = scheme([area(0, csch.A[i].S) for i in range(0, CORES)],
                          [area(0, csch.B[i].S) for i in range(0, CORES)],
                          area(0, csch.C.S), csch.mW, csch.mS)
                          
        ts_mcx.setScheme(wsch)
        for core in range(0, CORES):
            UA_Adata[core][w] = ts_mcx.A.core(core).utilization('A', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Bdata[core][w] = ts_mcx.A.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Cdata[core][w] = ts_mcx.A.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            
            UB_Bdata[core][w] = ts_mcx.B.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UB_Cdata[core][w] = ts_mcx.B.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        UAi = defaultdict()
        UAi['B'] = [UA_Bdata[core][w] for core in range(0, CORES)]
        UAi['C'] = [UA_Cdata[core][w] for core in range(0, CORES)]
        
        UAi = ts_mcx.secondaryInflateConst(UAi, minAperiods = minAperiods)
        for core in range(0, CORES):
            UA_Bdata[core][w] = UAi['B'][core]
            UA_Cdata[core][w] = UAi['C'][core]
        
        UC_Cdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        hdata[w] = ts_mcx.C.max_utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        Hdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
        
    
    #----------------------- Create LP Variables -----------------------------#
    problem = LpProblem("Hybrid",LpMinimize)
    
    UA_A = LpVariable.matrix("UA_A", range(CORES), lowBound = 0)
    UA_B = LpVariable.matrix("UA_B", range(CORES), lowBound = 0)
    UA_C = LpVariable.matrix("UA_C", range(CORES), lowBound = 0)
    
    UB_B = LpVariable.matrix("UB_B", range(CORES), lowBound = 0)
    UB_C = LpVariable.matrix("UB_C", range(CORES), lowBound = 0)
    
    UC_C = LpVariable("UC_C", lowBound = 0)
    
    h = LpVariable("h", lowBound = 0)
    H = LpVariable("H", lowBound = 0)
    
    Wa = LpVariable.matrix("Wa", range(CORES), lowBound = 0, cat=vartype)
    Wb = LpVariable.matrix("Wb", range(CORES), lowBound = 0, cat=vartype)
    Wc = LpVariable("Wc", lowBound = 0, cat=vartype);
    
    WO = LpVariable.matrix("WO", range(CORES), lowBound = 0)
    
    Wc.upBound = csch.mW
    
    
    #----------------------- Constraint Set 1 -----------------------------#
    for i, c in enumerate(Wa):
        Wa[i].upBound = csch.mW
        Wb[i].upBound = csch.mW
        WO[i].upBound = csch.mW
        
        if olap:
            problem += Wa[i] + Wc <= csch.mW
            problem += Wb[i] + Wc <= csch.mW
        else:
            problem += Wa[i] + Wb[i] + Wc <= csch.mW
            
        problem += WO[i] >= Wa[i] + Wb[i] + Wc - csch.mW
            
    #----------------------- Constraint Set 2 -----------------------------#
    for i in range(0, CORES):
        UA_A[i].upBound = 1
        UA_B[i].upBound = 1
        UA_C[i].upBound = 1
        
        UB_B[i].upBound = 1
        UB_C[i].upBound = 1
        
        if not len(ts_mcx.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts_mcx.A.core(i)])
        else:
            invminp = 0
            
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        IA_B = ohc.CPMD*Wb[i]*csch.B[i].S*invminp*tckinfw if olap else 0
        IA_C = ohc.CPMDavg*WO[i]*csch.B[i].S*invminp*tckinfa if olap else 0
        
        for j in range(0,csch.mW):
            C1 = (UA_Adata[i][j+1]-UA_Adata[i][j])
            C2 = UA_Adata[i][j]-j*C1
            problem += (UA_A[i]-C1*Wa[i] >= C2)
            C1 = (UA_Bdata[i][j+1]-UA_Bdata[i][j])
            C2 = UA_Bdata[i][j]-j*C1
            problem += (UA_B[i]-C1*Wa[i] >= C2 + IA_B)
            C1 = (UA_Cdata[i][j+1]-UA_Cdata[i][j])
            C2 = UA_Cdata[i][j]-j*C1
            problem += (UA_C[i]-C1*Wa[i] >= C2 + IA_C)
            
            C1 = (UB_Bdata[i][j+1]-UB_Bdata[i][j])
            C2 = UB_Bdata[i][j]-j*C1
            problem += (UB_B[i]-C1*Wb[i] >= C2)
            C1 = (UB_Cdata[i][j+1]-UB_Cdata[i][j])
            C2 = UB_Cdata[i][j]-j*C1
            problem += (UB_C[i]-C1*Wb[i] >= C2)
    
    UC_C.upBound = CORES
    for w in range(0, csch.mW):
        C1 = (UC_Cdata[w+1]-UC_Cdata[w])
        C2 = UC_Cdata[w]-w*C1
        problem += (UC_C-C1*Wc >= C2)
        
    #----------------------- Constraint Set 3 -----------------------------#
        problem += h >= hdata[w] + (Wc-w)*(hdata[w+1]-hdata[w])
        problem += H >= Hdata[w] + (Wc-w)*(Hdata[w+1]-Hdata[w])
        
    #----------------------- Constraint Set 4 -----------------------------#
    
    for c in range(0,CORES):
        problem += UA_A[c] <= 1
        problem += UA_B[c] + UB_B[c] <= 1
        problem += UA_B[c] + UB_B[c]
        
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_C <= CORES
        
    eps = 1e-6
    
    problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H <= CORES-eps
    
    problem += lpSum(UA_C) + lpSum(UB_C) + UC_C
    
    #----------------------- Solve -----------------------------#
    status = problem.solve(GUROBI(msg=0))

    if LpStatus[status] == 'Optimal':
        ps = scheme([area(Wa[i].varValue, csch.A[i].S) for i in range(0,CORES)],
                    [area(Wb[i].varValue, csch.B[i].S) for i in range(0,CORES)],
                    area(Wc.varValue, MAX_SETS), MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps
        
def LPLLCNapsack(ts_mcx, vartype, csch = bypass_scheme(), olap = True, debug = False, noIso = False):
    
    LWmax = 4
    Nbin = 16*LWmax
    
    ts_mcx.setScheme(csch)
    ts_mcx.A = ts_mcx.level('A')
    ts_mcx.B = ts_mcx.level('B')
    ts_mcx.C = ts_mcx.level('C')
    
    #----------------------- Gather Data -----------------------------#
    UA_Adata = np.zeros((CORES, csch.mW+1))
    UA_Bdata = np.zeros((CORES, csch.mW+1))
    UA_Cdata = np.zeros((CORES, csch.mW+1))
    
    UB_Bdata = np.zeros((CORES, csch.mW+1))
    UB_Cdata = np.zeros((CORES, csch.mW+1))
    UC_Cdata = np.zeros(csch.mW+1)
    uC_Cdata = np.zeros((len(ts_mcx.C), csch.mW+1))
    
    minAperiods = []
    for core in range(0, 4):
        if len(ts_mcx.A.core(core)) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in ts_mcx.A.core(core)]))
    
    for w in range(0,csch.mW+1):
        wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
                      [area(w, csch.B[i].S) for i in range(0, CORES)],
                      area(w, csch.C.S), csch.mW, csch.mS)
        ts_mcx.setScheme(wsch)
        for core in range(0, CORES):
            UA_Adata[core][w] = ts_mcx.A.core(core).utilization('A', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Bdata[core][w] = ts_mcx.A.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Cdata[core][w] = ts_mcx.A.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            
            UB_Bdata[core][w] = ts_mcx.B.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UB_Cdata[core][w] = ts_mcx.B.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        UAi = defaultdict()
        UAi['B'] = [UA_Bdata[core][w] for core in range(0, CORES)]
        UAi['C'] = [UA_Cdata[core][w] for core in range(0, CORES)]
        
        UAi = ts_mcx.secondaryInflateConst(UAi, minAperiods = minAperiods)
        for core in range(0, CORES):
            UA_Bdata[core][w] = UAi['B'][core]
            UA_Cdata[core][w] = UAi['C'][core]
        
        UC_Cdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        for t_i in range(len(ts_mcx.C)):
            uC_Cdata[t_i][w] = ts_mcx.C[t_i].utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiod = 0, mx_wss = capped_wss)
    
    
    #----------------------- Create LP Variables -----------------------------#
    problem = LpProblem("Hybrid",LpMinimize)
    
    UA_A = LpVariable.matrix("UA_A", range(CORES), lowBound = 0)
    UA_B = LpVariable.matrix("UA_B", range(CORES), lowBound = 0)
    UA_C = LpVariable.matrix("UA_C", range(CORES), lowBound = 0)
    
    UB_B = LpVariable.matrix("UB_B", range(CORES), lowBound = 0)
    UB_C = LpVariable.matrix("UB_C", range(CORES), lowBound = 0)
    
    UC_C = LpVariable("UC_C", lowBound = 0)
    
    uC_C = LpVariable.matrix("uC_C", range(len(ts_mcx.C)), lowBound = 0, cat='Continuous')
    
    L = LpVariable.matrix("L", range(len(ts_mcx.buffers)), 0, cat='Continuous')
    
    Wa = LpVariable.matrix("Wa", range(CORES), 0, csch.mW, cat='Integer')
    Wb = LpVariable.matrix("Wb", range(CORES), 0, csch.mW, cat='Integer')
    Wc = LpVariable("Wc", 0, csch.mW, cat='Integer')
    Wl = LpVariable("Wl", 0, LWmax, cat='Integer')
    
    WO = LpVariable.matrix("WO", range(CORES), 0, csch.mW)
    
    b_h = LpVariable("b_h", lowBound = 0)
    b_H = LpVariable("b_H", lowBound = 0)
    
    z = LpVariable.matrix("z", range(len(ts_mcx.C)), lowBound = 0)
    Z = LpVariable.matrix("Z", range(len(ts_mcx.C)), lowBound = 0)
    
    #----------------------- Constraint Set 1 -----------------------------#
        
    #----------------------- Constraint Set 2 -----------------------------#
    for i, c in enumerate(Wa):
        if olap:
            problem += Wa[i] + Wc + Wl <= csch.mW
            problem += Wb[i] + Wc + Wl <= csch.mW
        else:
            problem += Wa[i] + Wb[i] + Wc + Wl <= csch.mW
            
        problem += WO[i] >= Wa[i] + Wb[i] + Wc + Wl - csch.mW
            
    #----------------------- Constraint Set 3 -----------------------------#
    for q_i, q in enumerate(ts_mcx.buffers):
        problem += L[q_i] <= q.size_m
        
    assignment_sum = LpAffineExpression()
    for q_i, q in enumerate(ts_mcx.buffers):
        assignment_sum += L[q_i]*q.slots
    problem += assignment_sum <= 128*16*LWmax
    
    #----------------------- Constraint Set 4 -----------------------------#
    
    #Stuff needed in multiple places
    tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
    tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        
    #----------------------- Buffer Opt Expressions -----------------------#
    q2i_c = defaultdict()
        
    for q_i, q in enumerate(ts_mcx.buffers):
        q2i_c[q] = L[q_i]
        
    Dcache = defaultdict()
    
    for t_i, tsk in enumerate(ts_mcx):
        Dcache[tsk] = {'A' : LpAffineExpression(),
                       'B' : LpAffineExpression(),
                       'C' : LpAffineExpression()}
        for buff in tsk.R:
            if buff.tech == 'SBP': #copying
                Dcache[tsk]['A'] += ts_mcx.a_inflate_choice*q2i_c[buff]*(ohc.LINE_ACC_CBR   -ohc.LINE_ACC_CLCR   )/tsk.period,
                Dcache[tsk]['B'] +=                         q2i_c[buff]*(ohc.LINE_ACC_CBR   -ohc.LINE_ACC_CLCR   )/tsk.period,
                Dcache[tsk]['C'] +=                         q2i_c[buff]*(ohc.LINE_ACC_CBRavg-ohc.LINE_ACC_CLCRavg)/tsk.period
            else: #in-place
                Dcache[tsk]['A'] += ts_mcx.a_inflate_choice*(q2i_c[buff]/128.0)*ohc.CPMD   *tckinfw/tsk.period,
                Dcache[tsk]['B'] +=                         (q2i_c[buff]/128.0)*ohc.CPMD   *tckinfw/tsk.period,
                Dcache[tsk]['C'] +=                         (q2i_c[buff]/128.0)*ohc.CPMDavg*tckinfa/tsk.period
                
        for buff in tsk.W:
            if buff.tech == 'SBP': #copying
                Dcache[tsk]['A'] += ts_mcx.a_inflate_choice*q2i_c[buff]*(ohc.LINE_ACC_CBW   -ohc.LINE_ACC_CLCW   )/tsk.period,
                Dcache[tsk]['B'] +=                         q2i_c[buff]*(ohc.LINE_ACC_CBW   -ohc.LINE_ACC_CLCW   )/tsk.period,
                Dcache[tsk]['C'] +=                         q2i_c[buff]*(ohc.LINE_ACC_CBWavg-ohc.LINE_ACC_CLCWavg)/tsk.period
            else: #in-place
                Dcache[tsk]['A'] += ts_mcx.a_inflate_choice*(q2i_c[buff]/128.0)*ohc.CPMD   *tckinfw/tsk.period,
                Dcache[tsk]['B'] +=                         (q2i_c[buff]/128.0)*ohc.CPMD   *tckinfw/tsk.period,
                Dcache[tsk]['C'] +=                         (q2i_c[buff]/128.0)*ohc.CPMDavg*tckinfa/tsk.period
        
    #----------------------- Constraint Set 5 -----------------------------#
    for i in range(0, CORES):
        UA_A[i].upBound = 1
        UA_B[i].upBound = 1
        UA_C[i].upBound = 1
        
        UB_B[i].upBound = 1
        UB_C[i].upBound = 1
        
        if not len(ts_mcx.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts_mcx.A.core(i)])
        else:
            invminp = 0
            
        IA_B = ohc.CPMD*Wb[i]*csch.B[i].S*invminp*tckinfw if olap else 0
        IA_C = ohc.CPMDavg*WO[i]*csch.B[i].S*invminp*tckinfa if olap else 0
        
        for j in range(0,csch.mW):
            C1 = (UA_Adata[i][j+1]-UA_Adata[i][j])
            C2 = UA_Adata[i][j]-j*C1
            problem += (UA_A[i]-C1*Wa[i] >= C2)
            C1 = (UA_Bdata[i][j+1]-UA_Bdata[i][j])
            C2 = UA_Bdata[i][j]-j*C1
            problem += (UA_B[i]-C1*Wa[i] >= C2 + IA_B)
            C1 = (UA_Cdata[i][j+1]-UA_Cdata[i][j])
            C2 = UA_Cdata[i][j]-j*C1
            problem += (UA_C[i]-C1*Wa[i] >= C2 + IA_C)
            
            C1 = (UB_Bdata[i][j+1]-UB_Bdata[i][j])
            C2 = UB_Bdata[i][j]-j*C1
            problem += (UB_B[i]-C1*Wb[i] >= C2)
            C1 = (UB_Cdata[i][j+1]-UB_Cdata[i][j])
            C2 = UB_Cdata[i][j]-j*C1
            problem += (UB_C[i]-C1*Wb[i] >= C2)
    
    UC_C.upBound = CORES
    for w in range(0, csch.mW):
        C1 = (UC_Cdata[w+1]-UC_Cdata[w])
        C2 = UC_Cdata[w]-w*C1
        problem += (UC_C-C1*Wc >= C2)
        
    for t_i, tsk in enumerate(ts_mcx.C):
        tailC = - Dcache[tsk]['C']
        for w in range(csch.mW):
            C1 = uC_Cdata[t_i][w+1]-uC_Cdata[t_i][w]
            C2 = uC_Cdata[t_i][w]-w*C1
            problem += uC_C[t_i]-C1*Wc >= C2 + tailC
        
    #----------------------- Constraint Set 6 -----------------------------#
    h = b_h+lpSum(z)
    H = 3*b_H+lpSum(Z)
        
    for t_i in range(len(ts_mcx.C)):
        problem += z[t_i] >= uC_C[t_i]-b_h
        problem += Z[t_i] >= uC_C[t_i]-b_H
    
    #----------------------- Constraint Set 7 -----------------------------#
    DA_C = LpAffineExpression()
    DB_C = LpAffineExpression()
    DA_A = [LpAffineExpression() for p in range(4)]
    DA_B = [LpAffineExpression() for p in range(4)]
    DA_C = [LpAffineExpression() for p in range(4)]
    DB_B = [LpAffineExpression() for p in range(4)]
    DB_C = [LpAffineExpression() for p in range(4)]
    for c in range(0,CORES):
        for tsk in ts_mcx.A.core(c):
            DA_A[c] += Dcache[tsk]['A']
            DA_B[c] += Dcache[tsk]['B']
            DA_C += Dcache[tsk]['C']
        for tsk in ts_mcx.B.core(c):
            DB_B[c] += Dcache[tsk]['B']
            DB_C += Dcache[tsk]['C']
        
        problem += UA_A[c] - DA_A[c] <= 1
        problem += UA_B[c] + UB_B[c] - DA_B[c] - DB_B[c] <= 1
    
    DC_C = LpAffineExpression()
    for tsk in ts_mcx.C:
        DC_C += Dcache[tsk]['C']
        
    problem += lpSum(UA_C) + lpSum(UB_C) - DA_C - DB_C + UC_C - DC_C <= CORES
    eps = 1e-6
    
    problem += lpSum(UA_C) + lpSum(UB_C) - DA_C - DB_C + (CORES-1)*h + H <= CORES-eps
    problem += lpSum(UA_C) + lpSum(UB_C) - DA_C - DB_C + UC_C - DC_C
    
    #----------------------- Solve -----------------------------#
    status = problem.solve(GUROBI(msg=0))
    
    if LpStatus[status] == 'Optimal':
        ps = scheme([area(Wa[i].varValue, csch.A[i].S) for i in range(0,CORES)],
                    [area(Wb[i].varValue, csch.B[i].S) for i in range(0,CORES)],
                    area(Wc.varValue, MAX_SETS), MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps
        
def LPLLCKitchenSink(ts_mcx, vartype, csch = bypass_scheme(), olap = True, debug = False):
    
    LWmax = 4
    Nbin = 16*LWmax
    
    ts_mcx.setScheme(csch)
    
    ts_mcx.A = ts_mcx.level('A')
    ts_mcx.B = ts_mcx.level('B')
    ts_mcx.C = ts_mcx.level('C')
     
    #----------------------- Gather Data -----------------------------#
    uA_Adata = np.zeros((len(ts_mcx.A), csch.mW+1))
    uA_Bdata = np.zeros((len(ts_mcx.A), csch.mW+1))
    uA_Cdata = np.zeros((len(ts_mcx.A), csch.mW+1))
    
    uB_Bdata = np.zeros((len(ts_mcx.B), csch.mW+1))
    uB_Cdata = np.zeros((len(ts_mcx.B), csch.mW+1))
    
    uC_Cdata = np.zeros((len(ts_mcx.C), csch.mW+1))
    
    minAperiods = []
    for core in range(0, 4):
        if len(ts_mcx.A.core(core)) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in ts_mcx.A.core(core)]))
    
    for w in range(0,csch.mW+1):
        wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
                      [area(w, csch.B[i].S) for i in range(0, CORES)],
                      area(w, csch.C.S), csch.mW, csch.mS)
        ts_mcx.setScheme(wsch)
        
        for t_i in range(len(ts_mcx.A)):
            uA_Adata[t_i][w] = ts_mcx.A[t_i].utilization('A', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiod = 0, mx_wss = capped_wss)
            uA_Bdata[t_i][w] = ts_mcx.A[t_i].utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiod = 0, mx_wss = capped_wss)
            uA_Cdata[t_i][w] = ts_mcx.A[t_i].utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiod = 0, mx_wss = capped_wss)
            
        for t_i in range(len(ts_mcx.B)):
            uB_Bdata[t_i][w] = ts_mcx.B[t_i].utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiod = 0, mx_wss = capped_wss)
            uB_Cdata[t_i][w] = ts_mcx.B[t_i].utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiod = 0, mx_wss = capped_wss)
            
        for t_i in range(len(ts_mcx.C)):
            uC_Cdata[t_i][w] = ts_mcx.C[t_i].utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiod = 0, mx_wss = capped_wss)
    
    #----------------------- Create LP Variables -----------------------------#
    problem = LpProblem("Hybrid",LpMinimize)
    
    cA = LpVariable.dicts("cA", [(t_i, p) for t_i in range(len(ts_mcx.A)) for p in range(CORES)], 0, 1, cat='Integer')
    cB = LpVariable.dicts("cB", [(t_i, p) for t_i in range(len(ts_mcx.B)) for p in range(CORES)], 0, 1, cat='Integer')
    
    uA_A = LpVariable.dicts("uA_A", [(t_i, p) for t_i in range(len(ts_mcx.A)) for p in range(CORES)], lowBound = 0, cat='Continuous')
    uA_B = LpVariable.dicts("uA_B", [(t_i, p) for t_i in range(len(ts_mcx.A)) for p in range(CORES)], lowBound = 0, cat='Continuous')
    uA_C = LpVariable.dicts("uA_C", [(t_i, p) for t_i in range(len(ts_mcx.A)) for p in range(CORES)], lowBound = 0, cat='Continuous')
    
    uB_B = LpVariable.dicts("uB_B", [(t_i, p) for t_i in range(len(ts_mcx.B)) for p in range(CORES)], lowBound = 0, cat='Continuous')
    uB_C = LpVariable.dicts("uB_C", [(t_i, p) for t_i in range(len(ts_mcx.B)) for p in range(CORES)], lowBound = 0, cat='Continuous')
    
    uC_C = LpVariable.matrix("uC_C", range(len(ts_mcx.C)), lowBound = 0, cat='Continuous')
    
    L = LpVariable.dicts("L", [(q_i, j) for q_i in range(len(ts_mcx.buffers)) for j in range(16*3)], 0, 1, cat='Integer')
    cS = LpVariable.matrix("cS", q_i for q_i in range(len(ts_mcx.buffersH)), 0, 1, cat='Integer') #1 if high-crit prod/cons has CE
    iso_c = LpVariable.matrix("iso_c", range(len(ts_mcx.buffersH)), lowBound = 0, cat='Integer') #1 if buffer is fully locked in cache
    i_p = LpVariable.matrix("i_p", range(len(ts_mcx.buffersH)), lowBound = 0, cat='Integer') #1 if high-crit prod/cons is used in-place
    
    Wa = LpVariable.matrix("Wa", range(CORES), 0, csch.mW, cat='Integer')
    Wb = LpVariable.matrix("Wb", range(CORES), 0, csch.mW, cat='Integer')
    Wc = LpVariable("Wc", 0, csch.mW, cat='Integer')
    Wl = LpVariable("Wm", 0, LWmax, cat='Integer')
    
    WO = LpVariable.matrix("WO", range(CORES), 0, csch.mW)
    
    b_h = LpVariable("b_h", lowBound = 0)
    b_H = LpVariable("b_H", lowBound = 0)
    
    z = LpVariable.matrix("z", range(len(ts_mcx.C)), lowBound = 0)
    Z = LpVariable.matrix("Z", range(len(ts_mcx.C)), lowBound = 0)
    
    #----------------------- Map Tasks to Buffer LP Variables -----------------------------#
    t2C = defaultdict()
    
    for t_i in range(len(ts_mcx.A)):
        t2C[ts_mcx.A[t_i]] = [cA[t_i, p] for p in range(CORES)]
    for t_i in range(len(ts_mcx.B)):
        t2C[ts_mcx.B[t_i]] = [cB[t_i, p] for p in range(CORES)]
    
    print "set 1"
    #----------------------- Constraint Set 1 -----------------------------#
    for t_i in range(len(ts_mcx.A)):
        assignment_sum = LpAffineExpression()
        for p in range(CORES):
            problem += cA[t_i,p] <= 1
            assignment_sum += cA[t_i,p]
        problem += assignment_sum == 1
    
    for t_i in range(len(ts_mcx.B)):
        assignment_sum = LpAffineExpression()
        for p in range(CORES):
            problem += cB[t_i,p] <= 1
            assignment_sum += cB[t_i,p]
        problem += assignment_sum == 1
        
    print "set 2"
    #----------------------- Constraint Set 2 -----------------------------#
    for i, c in enumerate(Wa):
        if olap:
            problem += Wa[i] + Wc + Wl <= csch.mW
            problem += Wb[i] + Wc + Wl <= csch.mW
        else:
            problem += Wa[i] + Wb[i] + Wc + Wl <= csch.mW
            
        problem += WO[i] >= Wa[i] + Wb[i] + Wc + Wl - csch.mW
            
    print "set 3"
    #----------------------- Constraint Set 3 -----------------------------#
    for q_i, q in enumerate(ts_mcx.buffers):
        problem += L[q_i] <= q.size_m
        
    assignment_sum = LpAffineExpression()
    for q_i, q in enumerate(ts_mcx.buffers):
        iso_c[q_i] <= L[q_i]/q.size_m
        assignment_sum += L[q_i]*q.slots
    problem += assignment_sum <= 128*16*LWmax
    
    print "set 4"
    #----------------------- Constraint Set 4 -----------------------------#
    for q_i, q in enumerate(ts_mcx.buffersH):
        for p in range(CORES):
            problem += cS[q_i] <= t2C[q.R][p] - t2C[q.W][p] + 1
            problem += cS[q_i] <= t2C[q.W][p] - t2C[q.R][p] + 1
        problem += iso_b[q_i] <= cS + iso_c[q_i]
        problem += iso_b[q_i] <= 1
    
    #----------------------- Buffer Opt Expressions -----------------------#
    b2iso_c = defaultdict()
    b2i_p = defaultdict()
    
    for q_i, q in enumerate(ts_mcx.buffersH):
        b2iso_c[q] = iso_c[q_i]
        b2i_p[q] = i_p[q_i]
    
    b2L = defaultdict()
    for q_i, q in enumerate(ts_mcx.buffers):
        b2L[q] = L[q_i]
        
    Dipl   = defaultdict() #in-place-locked
    Dcache = defaultdict()
    
#     L
#     cS    : 1 if high-crit prod/cons buffer has CE
#     iso_c : 1 if high-crit prod/cons buffer is fully locked in cache
#     i_p   : 1 if high-crit prod/cons buffer is used in-place
    
    #for copied, add copy phases, part from cache, part from banks
    #for not copied, subtract load time for locked
    for t_i, tsk in enumerate(ts_mcx):
        
        Dipl[tsk] = {'A' : LpAffineExpression(),
                     'B' : LpAffineExpression(),
                     'C' : LpAffineExpression()}
        
        Dcopy[tsk] = {'A' : LpAffineExpression(),
                      'B' : LpAffineExpression(),
                      'C' : LpAffineExpression()}
        
        for buff in tsk.R:
            if buff.R.level == buff.W.level: # may be in place
                if not buff.R.level == 'C': #H buffer
                    Dipl[tsk]['A'] += -i_p[buff]*ts_mcx.a_inflate_choice*buff.size_m*(ohc.LINE_ACC_CLCR   +ohc.CPMD   /128.0)/tsk.period
                    Dipl[tsk]['B'] += -i_p[buff]                        *buff.size_m*(ohc.LINE_ACC_CLCR   +ohc.CPMD   /128.0)/tsk.period
                    Dipl[tsk]['C'] += -i_p[buff]                        *buff.size_m*(ohc.LINE_ACC_CLCRavg+ohc.CPMDavg/128.0)/tsk.period
                else:
                    Dipl[tsk]['A'] += ts_mcx.a_inflate_choice*b2L[buff]*(ohc.CPMD   /128.0)/tsk.period
                    Dipl[tsk]['B'] +=                        *b2L[buff]*(ohc.CPMD   /128.0)/tsk.period
                    Dipl[tsk]['C'] +=                        *b2L[buff]*(ohc.CPMDavg/128.0)/tsk.period
            if not (buff.R.level == 'C' and buff.W.level == 'C'): #set copy phases for anything that MIGHT have a copy phase. Dipl compensates when it doesn't.
                numLock = b2L[buff]
                numBank = buff.size_m-b2L[buff]
                Dcopy[tsk]['A'] += ts_mcx.a_inflate_choice*(numLock*ohc.LINE_ACC_CLCR    + numBank*ohc.LINE_ACC_CBR   )/tsk.period,
                Dcopy[tsk]['B'] +=                         (numLock*ohc.LINE_ACC_CLCR    + numBank*ohc.LINE_ACC_CBR   )/tsk.period,
                Dcopy[tsk]['C'] +=                         (numLock*ohc.LINE_ACC_CLCRavg + numBank*ohc.LINE_ACC_CBRavg)/tsk.period
        
        for buff in tsk.W:
            if buff.R.level == buff.W.level: # may be in place
                if not buff.R.level == 'C': #H buffer
                    Dipl[tsk]['A'] += -i_p[buff]*ts_mcx.a_inflate_choice*buff.size_m*(ohc.LINE_ACC_CLCW   +ohc.CPMD   /128.0)/tsk.period
                    Dipl[tsk]['B'] += -i_p[buff]                        *buff.size_m*(ohc.LINE_ACC_CLCW   +ohc.CPMD   /128.0)/tsk.period
                    Dipl[tsk]['C'] += -i_p[buff]                        *buff.size_m*(ohc.LINE_ACC_CLCWavg+ohc.CPMDavg/128.0)/tsk.period
                else:
                    Dipl[tsk]['A'] += ts_mcx.a_inflate_choice*b2L[buff]*(ohc.CPMD   /128.0)/tsk.period
                    Dipl[tsk]['B'] +=                        *b2L[buff]*(ohc.CPMD   /128.0)/tsk.period
                    Dipl[tsk]['C'] +=                        *b2L[buff]*(ohc.CPMDavg/128.0)/tsk.period
            if not (buff.R.level == 'C' and buff.W.level == 'C'): #set copy phases for anything that MIGHT have a copy phase. Dipl compensates when it doesn't.
                numLock = b2L[buff]
                numBank = buff.size_m-b2L[buff]
                Dcopy[tsk]['A'] += ts_mcx.a_inflate_choice*(numLock*ohc.LINE_ACC_CLCW    + numBank*ohc.LINE_ACC_CBW   )/tsk.period,
                Dcopy[tsk]['B'] +=                         (numLock*ohc.LINE_ACC_CLCW    + numBank*ohc.LINE_ACC_CBW   )/tsk.period,
                Dcopy[tsk]['C'] +=                         (numLock*ohc.LINE_ACC_CLCWavg + numBank*ohc.LINE_ACC_CBWavg)/tsk.period
                
    print "set 5"
    #TODO ADD IN 
    #-Constant secondary inflation
    #----------------------- Constraint Set 5 -----------------------------#
    for t_i, tsk in enumerate(ts_mcx.C):
        tailC = Dipl[tsk]['C'] - Dcopy[tsk]['C']
        for w in range(csch.mW):
            C1 = uC_Cdata[t_i][w+1]-uC_Cdata[t_i][w]
            C2 = uC_Cdata[t_i][w]-w*C1
            problem += uC_C[t_i]-C1*Wc >= C2 + tailC
    tst = []
    for t_i, tsk in enumerate(ts_mcx.A):
        Dtsk_A = Dipl[tsk]['A'] + Dcopy[tsk]['A']
        Dtsk_B = Dipl[tsk]['B'] + Dcopy[tsk]['B']
        Dtsk_C = Dipl[tsk]['C'] + Dcopy[tsk]['C']
        tst.append([])
        for p in range(CORES):
            tst[t_i].append([])
            tailA = Dtsk_A - (1 - cA[t_i, p])*100#uA_Adata[t_i][0]
            tailB = Dtsk_B - (1 - cA[t_i, p])*100#uA_Bdata[t_i][0]
            tailC = Dtsk_C - (1 - cA[t_i, p])*100#uA_Cdata[t_i][0]
            for w in range(csch.mW):
                C1 = (uA_Adata[t_i][w+1]-uA_Adata[t_i][w])
                C2 = uA_Adata[t_i][w]-w*C1
                problem += uA_A[t_i, p]-C1*Wa[p] >= C2 + tailA
                tst[t_i][p].append(uA_A[t_i, p]-C1*Wa[p] >= C2 + tailA)
                C1 = (uA_Bdata[t_i][w+1]-uA_Bdata[t_i][w])
                C2 = uA_Bdata[t_i][w]-w*C1
                problem += uA_B[t_i, p]-C1*Wa[p] >= C2 + tailB
                C1 = (uA_Cdata[t_i][w+1]-uA_Cdata[t_i][w])
                C2 = uA_Cdata[t_i][w]-w*C1
                problem += uA_C[t_i, p]-C1*Wa[p] >= C2 + tailC
    for t_i, tsk in enumerate(ts_mcx.B):
        Dtsk_B = Dipl[tsk]['B'] + Dcopy[tsk]['B']
        Dtsk_C = Dipl[tsk]['C'] + Dcopy[tsk]['C']
        for p in range(CORES):
            tailB = Dtsk_B - (1 - cB[t_i, p])*100#uB_Bdata[t_i][0]
            tailC = Dtsk_C - (1 - cB[t_i, p])*100#uB_Cdata[t_i][0]
            for w in range(csch.mW):
                C1 = uB_Bdata[t_i][w+1]-uB_Bdata[t_i][w]
                C2 = uB_Bdata[t_i][w]-w*C1
                problem += uB_B[t_i, p]-C1*Wb[p] >= C2 + tailB
                C1 = uB_Cdata[t_i][w+1]-uB_Cdata[t_i][w]
                C2 = uB_Cdata[t_i][w]-w*C1
                problem += uB_C[t_i, p]-C1*Wb[p] >= C2 + tailC
        
    print "set 6"
    #----------------------- Constraint Set 6 -----------------------------#
    h = b_h+lpSum(z)
    H = 3*b_H+lpSum(Z)
        
    #TODO: find indices of C tasks that correspond to v['C']['C'] indices
    for t_i in range(len(ts_mcx.C)):
        problem += z[t_i] >= uC_C[t_i]-b_h
        problem += Z[t_i] >= uC_C[t_i]-b_H
    
    print "set 7"
    #----------------------- Constraint Set 7 -----------------------------#
    UA_A = [LpAffineExpression() for p in range(CORES)]
    UA_B = [LpAffineExpression() for p in range(CORES)]
    UA_C = LpAffineExpression()
    
    UB_B = [LpAffineExpression() for p in range(CORES)]
    UB_C = LpAffineExpression()
    
    UC_C = LpAffineExpression()
    
    for p in range(0,CORES):
        if not len(ts_mcx.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts_mcx.A.core(i)])
        else:
            invminp = 0
            
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        IA_B = ohc.CPMD*Wb[i]*csch.B[i].S*invminp*tckinfw if olap else 0
        IA_C = ohc.CPMDavg*WO[i]*csch.B[i].S*invminp*tckinfa if olap else 0
        
        for t_i in range(len(ts_mcx.A)):
            UA_A[p] += uA_A[t_i, p]
            UA_B[p] += uA_B[t_i, p]
            UA_C += uA_C[t_i, p]
        UA_B[p] += IA_B
        UA_C += IA_C
        
        for t_i in range(len(ts_mcx.B)):
            UB_B[p] += uB_B[t_i, p]
            UB_C += uB_C[t_i, p]
        
        problem += UA_A[p] <= 1
        problem += UA_B[p] + UB_B[p] <= 1
    
    for t_i in range(len(ts_mcx.C)):
        UC_C += uC_C[t_i]
        
    problem += UA_C + UB_C + UC_C <= CORES
    eps = 1e-6
    
    problem += UA_C + UB_C + (CORES-1)*h + H <= CORES-eps
    problem += UA_C + UB_C + UC_C
    
#     print problem
    print "solving"
    #----------------------- Solve -----------------------------#
    status = problem.solve(GUROBI(msg=0))
    
    if LpStatus[status] == 'Optimal':
        ps = scheme([area(Wa[i].varValue, csch.A[i].S) for i in range(0,CORES)],
                    [area(Wb[i].varValue, csch.B[i].S) for i in range(0,CORES)],
                    area(Wc.varValue, MAX_SETS), MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps