#!/usr/bin/env python

from __future__ import division

import matplotlib
matplotlib.use('Agg')
import matplotlib.pyplot as plt
LINE_STYLE = ['b:+', 'g-', 'r-s', 'c--', 'k-.']

import traceback
import argparse
import random
import time
import numpy as np
from collections import defaultdict
from csv import DictWriter
from util.stats import proportion_ci

from multiprocessing import Pool, cpu_count
from schedcat.util.storage import storage
from generator.generator import DesignPointGenerator

from distributions import *

from schedcat.model.LLC_params import *
from schedcat.generator.mcx_tasks import TaskGeneratorMCX
from LLCOptimizer import *
from LLCCanvas import LLCCanvas
from DAGGenerator import *
import schedcat.model.ohc as ohc
import loadohs
import graph

from BufferSizer import *

from sched_check import *

import multiprocessing
# We must import this explicitly, it is not imported by the top-level
# multiprocessing module.
import multiprocessing.pool
from random import randint

MIN_SAMPLES = 100
MAX_SAMPLES = 2000 #2000
MAX_CI      = 0.05 #.05
CONFIDENCE  = 0.95 #.95

topdir = "~/LLC_Code/MCXSchedStudy/"

def makeDependencies(ts, dp):
    # swap the squeeze task into random position
    shuf = random.randint(0, len(ts)-1)
    ts[-1], ts[shuf] = ts[shuf], ts[-1]
    
    nrTasks = len(ts)
    subtasksets = []
    count = 0
    
    while count < len(ts):
        compSize = dp.num_graphs()
        subts = ts[count:count+compSize]
        subtasksets.append(subts)
        count += compSize
    
    # nrGraphs = min(dp.num_graphs(), nrTasks)
# 
#     shares = []
#     for i in range(nrGraphs):
#         shares.append(1.0 - random.random())
#     
#     weight = sum(shares)
#     shares = [int((s/weight)*nrTasks + 0.5) for s in shares]
# 
#     # we may have gained/lost a node due to rounding
#     # add/remove node from any share with space
#     todrop = sum(shares) - nrTasks
#     if todrop > 0:
#         for i in range(todrop):
#             candidates = [i for i,y in enumerate(shares) if y > 1]
#             shares[random.choice(candidates)] -= 1
#     elif todrop < 0:
#         for i in range(-1*todrop):
#             shares[random.randint(0,len(shares)-1)] += 1
# 
#     # make sure that no graph has zero nodes
#     # steal from graphs at random
#     nullGraphs = [i for i,y in enumerate(shares) if y == 0]
#     while nullGraphs:
#         stealGraphs = [i for i,y in enumerate(shares) if y > 1]
#         assert stealGraphs
#         shares[random.choice(stealGraphs)] -= 1
#         shares[nullGraphs[-1]] += 1
#         nullGraphs.pop()
# 
#     assert sum(shares) == nrTasks
# 
#     subtasksets = []
#     count = 0
#     for i in range(nrGraphs):
#         subts = ts[count:count+shares[i]]
#         assert len(subts) > 0
#         subtasksets.append(subts)
#         count += shares[i]
# 
#     print len(ts), [len(subtaskset) for subtaskset in subtasksets]

    for tsk in ts:
        tsk.R = []
        tsk.W = []
    graphs = []
    for subts in subtasksets:
        graphsz = len(subts)
        min_depth = max(1, int(dp.depth_factor[0] * graphsz))
        max_depth = max(1, int(dp.depth_factor[1] * graphsz))
        gg = graph.GraphGenerator(tasks.uniform_int(graphsz, graphsz),
                                 tasks.uniform_int(min_depth, max_depth),
                                 dp.node_placement,
                                 dp.fan_out, dp.fan_in_cap,
                                 dp.edge_distance,
                                 dp.nr_source, dp.nr_sink,
                                 False, False)
        g = gg.graph()
                                 
        assert len(g.nodes) == graphsz
        
        # bind tasks/nodes, and compute node wss
        for i in range(graphsz):
            g.nodes[i].task = subts[i]
            subts[i].node = g.nodes[i]
            
            for e in g.nodes[i].inEdges:
                e.R = subts[i]
            for e in g.nodes[i].outEdges:
                e.W = subts[i]
                
            subts[i].R = g.nodes[i].inEdges
            subts[i].W = g.nodes[i].outEdges
            subts[i].graph = g
            # task working set is the sum of inputs
#             subts[i].wss = 0 if len(g.nodes[i].inEdges) == 0 else sum([w.wss for w in g.nodes[i].inEdges])

        # assign working sets to each edge
        #note: msz refers to percent of load time
        for i in range(graphsz):
            subts[i].msz = 128*dp.msz[subts[i].level]()*subts[i].wss
        
#         print [subts[i].msz for i in range(graphsz)]
#         print [e.size_m for i in range(len(g.edges))]
            
        # single root
        root = [r for r in g.nodes if r.isSrc][0]

        # adjust every task to have same execution rate as the source. must
        # adjust both period and execution time. utilization remains unchanged.
        
        subA = [t for t in subts if t.level == 'A']
        subB = [t for t in subts if t.level == 'B']
        subC = [t for t in subts if t.level == 'C']
        
        if len(subA) > 0:
            aPeriod = subA[0].period
            for t in subA:
                t.adjust_to_new_period(aPeriod)
                t.period = aPeriod
                t.deadline = aPeriod
        
        if len(subB) > 0:
            bPeriod = subB[0].period
            for t in subB:
                t.adjust_to_new_period(bPeriod)
                t.period = bPeriod
                t.deadline = bPeriod
                
        if len(subC) > 0:
            cPeriod = subC[0].period
            for t in subC:
                t.adjust_to_new_period(cPeriod)
                t.period = cPeriod
                t.deadline = cPeriod
            
#         for t in subts:
#             if root.task != t:
#                 t.adjust_to_new_period(root.task.period)
#                 t.period = root.task.period
#                 t.deadline = t.period
#                 t.cost = int(t.period * tutil + 0.5)

        graphs.append(g)

    
    # tag each graph with an id
    for i,g in enumerate(graphs):
        g.id = i

    return ts, graphs, subtasksets

# def getCritConnected(tsk, dag):
# 
#     dag.append(tsk)
#     
#     if hasattr(tsk, 'R'):
#         for buff in tsk.R:
#             if buff.W.level == tsk.level and (not buff.W in dag):
#                 getCritConnected(buff.W, dag)
#                 
#     if hasattr(tsk, 'W'):
#         for buff in tsk.W:
#             if buff.R.level == tsk.level and (not buff.R in dag):
#                 getCritConnected(buff.R, dag)

# def getABconnected(tsk, dag):
# 
#     dag.append(tsk)
#     
#     if hasattr(tsk, 'R'):
#         for buff in tsk.R:
#             if (not buff.W.level == 'C') and (not buff.W in dag):
#                 getABconnected(buff.W, dag)
#                 
#     if hasattr(tsk, 'W'):
#         for buff in tsk.W:
#             if (not buff.R.level == 'C') and (not buff.R in dag):
#                 getABconnected(buff.R, dag)

def make_mcx_task_set(dp):
    tgMCX = TaskGeneratorMCX(
                             period = NAMED_PERIODS[dp.period],
                             util = NAMED_UTILIZATIONS[dp.task_util],
                             util_dist = NAMED_UTIL_DISTS[dp.util_dist],
                             mem_util = NAMED_MEMORY[dp.memory_util],
                             avg_percent = AVG_PERCENT,
                             a_inflation = NAMED_A_INFLATIONS[dp.inflation_A],
                             pfuncs = P)
    ts = tgMCX.make_task_set(max_util = dp.sys_util)
    
    ts.m = CORES
    
    tsA, graphsA, subtasksetsA = makeDependencies(ts.level('A'), dp)
    tsB, graphsB, subtasksetsB = makeDependencies(ts.level('B'), dp)
    tsC, graphsC, subtasksetsC = makeDependencies(ts.level('C'), dp)
    
    ts.subsets = {'A' : subtasksetsA, 'B' : subtasksetsB, 'C' : subtasksetsC}
    
    ts.buffers = []


    for g in graphsA:
        ts.buffers.extend(g.edges)
    for g in graphsB:
        ts.buffers.extend(g.edges)
    for g in graphsC:
        ts.buffers.extend(g.edges)
        
    ts.scheme = bypass_scheme()
    
    add_cross_crit_buffers(ts)
                
    LPSizeBuffers(ts)
    return ts

def complete(results, n):
    if n < MIN_SAMPLES:
        return False
    elif n > MAX_SAMPLES:
        return True
    else:
        for key in results:
            if proportion_ci(results[key], n, CONFIDENCE) > MAX_CI:
                return False
        return True

def update_mean(old_mean, n, new_sample):
    return (old_mean*n + new_sample)/(n+1)

def timeout(func, ts_mcx, vartype, csch = bypass_scheme(), olap = True, debug = False, noIso = False):
    '''This function will spwan a thread and run the given function using the args, kwargs and 
    return the given default value if the timeout_duration is exceeded 
    ''' 
    import threading
    class InterruptableThread(threading.Thread):
        def __init__(self):
            threading.Thread.__init__(self)
            self.result = 0
        def run(self):
            try:
#                 print "started with: "
                self.result = func(ts_mcx, vartype, csch, olap = olap, debug = debug, noIso = noIso)
#                 print "finished with", self.result
            except:
                self.result = 0
                
    it = InterruptableThread()
    it.start()
    it.join(10)
    
    if it.result == 0:
        print "didn't finish"
        return None
    else:
#         print "returning:", it.result
        return it.result

def process_dp(dp):
    dporig = Custom()
    
    dporig.cross_crit = dp.cross_crit
    dporig.num_graphs = dp.num_graphs
    dporig.depth_factor = dp.depth_factor
    dporig.node_placement = dp.node_placement
    dporig.fan_out = dp.fan_out
    dporig.edge_distance = dp.edge_distance
    dporig.msz = dp.msz

    # convert names to distributions
    dp.cross_crit = NAMED_CROSS_CRIT_COMMS[dp.cross_crit]
    dp.num_graphs = NAMED_NUM_GRAPHS[dp.num_graphs]
    dp.depth_factor = NAMED_HEIGHT_FACTORS[dp.depth_factor]
    dp.node_placement = NAMED_SHAPES[dp.node_placement]
    dp.fan_out = NAMED_FAN[dp.fan_out]
    dp.edge_distance = NAMED_EDGE_HOP[dp.edge_distance]
    dp.msz = NAMED_MESSAGE_SIZES[dp.msz]

    # slam in unchaging values
    dp.nr_source = graph.uniform(1,1)
    dp.nr_sink = graph.uniform(1,1)
    
    
    sched_res = defaultdict(float)
    count = defaultdict(int)
    ps_hist = defaultdict(float) #np.zeros(51)
    
    n = 0
    
    while not complete(sched_res, n):
        ts = make_mcx_task_set(dp)
        unmnwts = []
        loadwtsAB = []
        loadwtsC = []
        
        for tsk in ts:
            unmnwts.append(unmanwt())
            loadwtsAB.append(loadedinfAB())
            
        for tsk in ts.level('C'):
                loadwtsC.append(loadedinfC())
        
        ts.useCshared()
        ts.setLLCAlloc(bypass_scheme())
        ohc.SetOheads(ts, 'mc2') 
        
        #---------inflate for loads in MC-Cshared------------#
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsAB[idx]    
                        
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] *= loadwtsC[idx]  
        #------------------------------------------------#
        
        #NAIVE#
        #----------inflate for AB bank interference for unmanaged buffers-------------#

        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    if tsk.level is 'A' or tsk.level is 'B':
                        for crt in {'A', 'B', 'C'}:
                            tsk.cost[crt][s][w] *= unmnwts[idx]
        
        ts.coreAssign(useMC2 = True)
        
        ts.applyNaive()
        ts.add_buffer_copy_times()
        
        psNaive = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = True, debug = False, noIso = True)
        if psNaive == None:
            psNaive = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = False, debug = False, noIso = True)
        schedNaive = psNaive != None
        ts.remove_buffer_copy_times()
        #---------undo bank interference------------#
        
        for idx, tsk in enumerate(ts):
            for s in range(0, 17):
                for w in range(0, 17):
                    if tsk.level is 'A' or tsk.level is 'B':
                        for crt in {'A', 'B', 'C'}:
                            tsk.cost[crt][s][w] /= unmnwts[idx]
        
        #Enable CB, assign to C banks# 
        ts.coreAssign(useMC2 = True)
        
        #---------add buffer copy times-------------#
        
        ts.applySBP()
  
        ts.setLLCAlloc(bypass_scheme())
        ts.add_buffer_copy_times()
        
        psSBP = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = True, debug = False)
        if psSBP== None:
            psSBP = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = False, debug = False)
        schedSBP = psSBP != None 
        ts.setLLCAlloc(bypass_scheme())
        
        ts.remove_buffer_copy_times()
        
        ts.applyCE()
        
        ts.add_buffer_copy_times()
        psCEnoDAG = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = True, debug = False)
        if psCEnoDAG == None:
            psCEnoDAG = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = False, debug = False)
        schedCEnoDAG = psCEnoDAG != None 

        ts.remove_buffer_copy_times()
        
        ts.applyNaive()
        ts.applySBP()

        ts.dagAssign('A')
        ts.dagAssign('B')

        ts.applyCE()

        #--------add changed buffer times-----------#
        ts.add_buffer_copy_times()
  
        psCE = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = True, debug = False)
        if psCE== None:
            psCE = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = False, debug = False)
        schedCE = psCE != None
        
        if (not schedCE) and (schedCEnoDAG):
            ts.setLLCAlloc(bypass_scheme())
            ts.remove_buffer_copy_times()
            ts.coreAssign(useMC2 = True)
            ts.applyNaive()
            ts.applySBP()
            ts.applyCE()
            ts.add_buffer_copy_times()
            psCL = psCEnoDAG
        
        schedCE = schedCEnoDAG or schedCE
        
        #Lock buffers#
        
        psCL = timeout(LPLLCNapsack, ts, 'Integer', csch = default_scheme(), olap = True, debug = False)
        if psCL== None:
            psCL = timeout(LPLLCNapsack, ts, 'Integer', csch = default_scheme(), olap = False, debug = False)
#         
        schedCL = psCL != None
            
        #Infinite secondary cache#
        
        ts.remove_buffer_copy_times() 
        
        ts.applyIdeal()
        
        ts.add_buffer_copy_times()
        ts.add_buffer_lock_benefits()
        
        psIdeal = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = True, debug = False)
        if psIdeal== None:
            psIdeal = timeout(LPLLCways, ts, 'Integer', csch = default_scheme(), olap = False, debug = False)
        
        schedIdeal = psIdeal != None  
        
        ts.remove_buffer_copy_times()
        ts.remove_buffer_lock_benefits()
        ###UEDF###
        for idx, tsk in enumerate(ts):
            if tsk.level is 'C':
                break
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsAB[idx]    
                        
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    for crt in {'A', 'B', 'C'}:
                        tsk.cost[crt][s][w] /= loadwtsC[idx] 
        
        ts.useCiso()
        
        for tsk in ts:
            for s in range(0, 17):
                for w in range(0, 17):
                    tsk.cost['B'][s][w] = tsk.cost['A'][s][w]   
        for idx, tsk in enumerate(ts.level('C')):
            for s in range(0, 17):
                for w in range(0, 17):
                    tsk.cost['B'][s][w] /= unmnwts[idx]
        
        for tsk in ts:
            tsk.core = 0
            tsk.level = 'B'
        
        ts.setLLCAlloc(uni_scheme())
        ohc.SetOheads(ts, 'pedf')
        mx_wss_all = max([tsk.wss for tsk in ts])
        mx_wss = {'B' : [mx_wss_all, mx_wss_all, mx_wss_all, mx_wss_all],
                  'C' : 0}
        ohc.CPMD = 30
        schedUEDF = ts.utilization('B', approximate = False, inflated = True, c_taskset = [], minAperiods = [0, 0, 0, 0], mx_wss = mx_wss) <= 1
        
        try:
            assert (not schedSBP) or schedCE
            assert (not schedCE) or schedCL
            assert (not schedCL) or schedIdeal
        except AssertionError:
            _, _, tb = sys.exc_info()
            tb_info = traceback.extract_tb(tb)
            _, _, _, text = tb_info[-1]

            print "Assertion error", text
            print "dp: ", dp
#             print "utils:", schedCLdata, schedIdealdata 
#             raise
        sched_res['U-EDF']     = update_mean(sched_res['U-EDF']    , n, schedUEDF)
        sched_res['Naive']     = update_mean(sched_res['Naive']    , n, schedNaive)
        sched_res['SBP']       = update_mean(sched_res['SBP']      , n, schedSBP)
        sched_res['SBP+CE']    = update_mean(sched_res['SBP+CE']   , n, schedCE)
        sched_res['SBP+CE+CL'] = update_mean(sched_res['SBP+CE+CL'], n, schedCL)
        sched_res['Ideal']     = update_mean(sched_res['Ideal']    , n, schedIdeal)
        
        count['Count'] = n+1

        n+=1
    
    dp.cross_crit = dporig.cross_crit
    dp.num_graphs = dporig.num_graphs
    dp.depth_factor = dporig.depth_factor
    dp.node_placement = dporig.node_placement
    dp.fan_out = dporig.fan_out
    dp.edge_distance = dporig.edge_distance
    dp.msz = dporig.msz
    
    del dp.nr_source
    del dp.nr_sink
    
    return dict(dp.items() + sched_res.items()+count.items())
    

def myrange(start, end, inc):
    return np.arange(start, end+inc, inc)
    
def valid(dp):
    return True

class NoDaemonProcess(multiprocessing.Process):
    # make 'daemon' attribute always return False
    def _get_daemon(self):
        return False
    def _set_daemon(self, value):
        pass
    daemon = property(_get_daemon, _set_daemon)
    
# We sub-class multiprocessing.pool.Pool instead of multiprocessing.Pool
# because the latter is only a wrapper function, not a proper class.
class MyPool(multiprocessing.pool.Pool):
    Process = NoDaemonProcess

def sleepawhile(t):
    print("Sleeping %i seconds..." % t)
    time.sleep(t)
    return t

def work(num_procs):
    print("Creating %i (daemon) workers and jobs in child." % num_procs)
    pool = multiprocessing.Pool(num_procs)

    result = pool.map(sleepawhile,
        [randint(1, 5) for x in range(num_procs)])

    # The following is not really needed, since the (daemon) workers of the
    # child's pool are killed when the child is terminated, but it's good
    # practice to cleanup after ourselves anyway.
    pool.close()
    pool.join()
    return result

def test():
    print("Creating 5 (non-daemon) workers and jobs in main process.")
    pool = MyPool(5)

    result = pool.map(work, [randint(1, 5) for x in range(5)])

    pool.close()
    pool.join()
    print(result)

def thread_tests(args, outs, design_points):
    pool = MyPool(processes = args.processors)
    try:
        for i, row in enumerate(pool.imap_unordered(process_dp, design_points)):
            if not row:
                continue
            sys.stderr.write('\rdone {0:%}'.format(i/len(design_points)))
#             print '\rdone {0:%}'.format(i/len(design_points))
#             print row
            title = titlefor(row)
            outs[title].writerow(row)
        pool.close()
    except Exception as e:
        print "hit an exception"
        print e
        pool.terminate()
        pool.join()

def titlefor(headers):
    return headers['util_dist'] + "--" + headers['period'] + "--" + headers['task_util'] + "--" + headers['memory_util'] + "--" + headers['msz'] + "--" + headers['num_graphs']

def main():
    parser = argparse.ArgumentParser()
    # parser.add_argument('-d', "--datafile", type = argparse.FileType('w'),
                        # default = sys.stdout,
                        # help = "File to output csv file to")
    parser.add_argument('-p', "--plot", action='store_true',
                        help = "For plotting generated PETs")
    parser.add_argument('-m', "--processors", default=cpu_count(), type = int,
                        help = "Number of processors to execute on")

    args = parser.parse_args()
    exp = storage()

    exp.task_util = NAMED_UTILIZATIONS.keys()
    exp.inflation_A = NAMED_A_INFLATIONS.keys()
    exp.period = NAMED_PERIODS.keys()
    exp.memory_util = NAMED_MEMORY.keys()
    exp.util_dist = NAMED_UTIL_DISTS.keys()
    exp.sys_util = myrange(0.1, 9.1, 0.2)
    
    #graph stuff
    exp.num_graphs =     NAMED_NUM_GRAPHS.keys()
    exp.msz =            NAMED_MESSAGE_SIZES.keys()
    
    exp.depth_factor =   ['combined']#, 'uni-medium', 'uni-tall', 'pipeline']
    exp.node_placement = ['binomial']
    exp.fan_out =        ['uniform_3']#, 'none']
    exp.edge_distance =  ['none']#, geometric_3']#
    exp.fan_in_cap =     [3]
    exp.cross_crit = NAMED_CROSS_CRIT_COMMS.keys()
    
    design_points = [dp for dp in DesignPointGenerator(exp, is_valid = valid)]

    random.seed(43215)
    # random.shuffle(design_points)
    
    loadohs.load("../oheads/")
    
    print "Total design points: ", len(design_points)
    # if(len(os.listdir(os.getcwd() + "/results/")) > 0):
        # print "already have results"
        # exit()
    
    outs = dict()
    for dp in design_points:
        title = titlefor(dp)
        if not title in outs:
            outs[title] = DictWriter(open("../results_online/" + title, 'w'), exp.keys()+['U-EDF', 'Naive', 'SBP', 'SBP+CE', 'SBP+CE+CL', 'Ideal', 'Count'])
            outs[title].writeheader()

#     for i, t in enumerate(design_points):
#         process_dp(design_points[i])

    thread_tests(args, outs, design_points)

if __name__ == '__main__':
    main()